# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
import json
import logging
from datetime import datetime

from awsiot.greengrasscoreipc.model import (
    PublishToIoTCoreRequest
)

from feedbackloop.service.IPCUtils import QOS_TYPE, TIMEOUT

DEFAULT_PREDICTION_TOPIC = "feedback/message/prediction"

class ModelPredictionPublisher(object):
    def __init__(self, ipc_client):
        self._client = ipc_client

    def publish(self, config_id, model_prediction, model_input_location, metadata, prediction_topic, timestamp=datetime.now()):
        '''
        Publishes prediction to IoT Core topic prediction_topic through MQTT.
        Also logs timestamp of prediction upload in metadata.
        '''
        request = PublishToIoTCoreRequest()
        request.topic_name = prediction_topic
        request.qos = QOS_TYPE

        extra_metadata = {}

        if timestamp:
            extra_metadata["publish-ts"] = str(timestamp)

        metadata.update(extra_metadata)

        request.payload = bytes(json.dumps({
                "source-ref": model_input_location.uri,
                "model-prediction": model_prediction,
                "metadata": metadata,
                "config-id": config_id
            }), "utf-8")

        operation = self._client.new_publish_to_iot_core()
        operation.activate(request)
        future = operation.get_response()
        future.result(TIMEOUT)

        logging.info("Published model prediction for S3 object: {}, to MQTT topic: {} for feedback config-id: {}".format(
            model_input_location.uri, prediction_topic, config_id
        ))



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
import six
from feedbackloop.model.s3_key import S3Key

class S3Location(object):
    def __init__(self, bucket, key):
        if not isinstance(bucket, six.string_types):
            raise ValueError("Failed to create S3Location. Parameter 'bucket' must be of type string.")

        if not isinstance(key, S3Key):
            raise ValueError("Failed to create S3Location. Parameter 'key' must be of type S3Key.")

        self._bucket = bucket
        self._key = key

    @property
    def bucket(self):
        return self._bucket
    
    @property
    def key(self):
        return self._key
    
    @property
    def uri(self):
        return "s3://{}/{}".format(self._bucket, self._key.key)


# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
import logging
import uuid
from threading import Thread
from feedbackloop.model.filename import Filename
from feedbackloop.model.s3_key import S3Key
from feedbackloop.model.s3_location import S3Location
from feedbackloop.service.status_publisher import StatusMessageSuccess
from feedbackloop.service.status_publisher import StatusMessageSampleDropped
from feedbackloop.service.status_publisher import StatusMessageError

COUNTER_SAMPLE_NOT_USED = "SampleNotUsed"
COUNTER_PUBLISHED_SAMPLE_AND_PREDICTION = "SampleAndPredictionPublished"


class RequestProcessor(Thread):
    '''
        Responsible for handling the resource intensive aspects of feedback upload
        requests (S3 upload and model prediction publishing)
    '''
    def __init__(
        self,
        request_queue,
        metrics_publisher,
        status_publisher,
        model_input_uploader,
        model_prediction_publisher
    ):
        Thread.__init__(self)
        self._request_queue = request_queue
        self._metrics_publisher = metrics_publisher
        self._status_publisher = status_publisher
        self._model_input_uploader = model_input_uploader
        self._model_prediction_publisher = model_prediction_publisher

    def _handle(self, feedback_input, feedback_config):
        sampling_strategy = feedback_config.sampling_strategy

        if feedback_input.filename:
            filename = Filename(feedback_input.filename, feedback_config.file_ext)
        else:
            # If no filename argument is given to S3Key, a UUID will be used
            filename = Filename(str(uuid.uuid4()), feedback_config.file_ext)
        s3_key = S3Key(prefix=feedback_config.s3_prefix, filename=filename)
        model_input_upload_location = S3Location(feedback_config.s3_bucket, s3_key)
        if not sampling_strategy or sampling_strategy.should_use_sample(feedback_input.model_prediction):
            try:
                upload_result = self._model_input_uploader.upload(
                    s3_location=model_input_upload_location,
                    model_input=feedback_input.model_input,
                    content_type=feedback_config.content_type,
                    metadata=feedback_input.metadata
                )
            except Exception as e:
                error_str = "Failed to upload model input data due to exception. Model prediction will not be published. Exception type: {}, error: {}".format(
                 type(e).__name__, e
                )
                logging.warn(error_str)
                self._status_publisher.publish(StatusMessageError(exception=e, message=error_str), feedback_input.status_topic)
            else:
                self._model_prediction_publisher.publish(
                    config_id=feedback_input.config_id,
                    model_prediction=feedback_input.model_prediction,
                    model_input_location=model_input_upload_location,
                    metadata=feedback_input.metadata,
                    prediction_topic=feedback_input.prediction_topic
                )
                self._metrics_publisher.publish_counter(COUNTER_PUBLISHED_SAMPLE_AND_PREDICTION, 1)
                success_str = "Successfully uploaded model input and published model prediction."
                logging.info(success_str)
                self._status_publisher.publish(StatusMessageSuccess(s3_response_json=upload_result), feedback_input.status_topic)
        else:
            logging.info("Model input and model prediction were dropped by the configured sampling strategy for feedback configuration with id: {}".format(feedback_input.config_id))
            self._metrics_publisher.publish_counter(COUNTER_SAMPLE_NOT_USED, 1)
            self._status_publisher.publish(StatusMessageSampleDropped(), feedback_input.status_topic)

    def _process_input(self):
        # Queue::get will default to blocking if the queue is empty
        feedback_input, feedback_config = self._request_queue.get()
        try:
            logging.info("Processing request.")
            self._handle(feedback_input, feedback_config)
        except Exception as e:
            logging.warn(
                "Unexpected exception raised while processing request. Exception type: {}, error: {}".format(
               type(e).__name__, e
            ))
        finally:
            # call task done to indicate to queue that request
            # has been processed.
            logging.info("Finished processing request.")
            self._request_queue.task_done()

    def run(self):
        while True:
            self._process_input()



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
import os


class S3Key(object):
    def __init__(self, prefix, filename):
        self._key = os.path.join(prefix.strip('/'), filename.filename)

    @property
    def key(self):
        return self._key



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
RESPONSE = "response"
RESPONSE_FIELD_STATUS = "status"
RESPONSE_FIELD_ERROR_MSG = "error_message"
RESPONSE_FIELD_ERROR_CLASS = "error"

VALUE_SUCCESS = "success"
VALUE_FAIL = "fail"


def generate_success_response(**kwargs):
 success_response = {
     RESPONSE: {
         RESPONSE_FIELD_STATUS: VALUE_SUCCESS
     }
 }

 if kwargs:
     success_response[RESPONSE].update(kwargs)

 return success_response


def generate_error_response(error_class, error_msg, **kwargs):
 error_response = {
     RESPONSE: {
         RESPONSE_FIELD_STATUS: VALUE_FAIL,
         RESPONSE_FIELD_ERROR_MSG: error_msg,
         RESPONSE_FIELD_ERROR_CLASS: error_class
     }
 }

 if kwargs:
     error_response[RESPONSE].update(kwargs)

 return error_response



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.

import logging
import awsiot.greengrasscoreipc

from awsiot.greengrasscoreipc.model import (
    ConfigurationUpdateEvents,
    GetConfigurationRequest,
    SubscribeToConfigurationUpdateRequest,
    QOS
)

from threading import Condition

FEEDBACKLOOP_COMPONENT = "com.example.FeedbackLoop"
QOS_TYPE = QOS.AT_LEAST_ONCE
TIMEOUT = 10

feedbackloop_update_condition = Condition()


class IPCUtils:
    def __init__(self, ipc_client):
        self.ipc_client = ipc_client

    def get_feedbackloop_configuration(self):
        r"""
        Ipc client creates a request and activates the operation to get the configuration of
        the feedbackloop configuration.
        :return: A dictionary object of DefaultConfiguration from the recipe.
        """
        try:
            request = GetConfigurationRequest(component_name=FEEDBACKLOOP_COMPONENT)
            operation = self.ipc_client.new_get_configuration()
            operation.activate(request).result(TIMEOUT)
            result = operation.get_response().result(TIMEOUT)
            return result.value
        except Exception as e:
            logging.error(
                "Exception occured during fetching the Feedback Loop configuration: {}".format(e)
            )
            exit(1)

    def get_feedbackloop_config_updates(self):
        r"""
        Ipc client creates a request and activates the operation to subscribe to the configuration changes.
        """
        try:
            subsreq = SubscribeToConfigurationUpdateRequest(component_name=FEEDBACKLOOP_COMPONENT)

            subscribe_operation = self.ipc_client.new_subscribe_to_configuration_update(
                FeedbackLoopConfigUpdateHandler()
            )
            subscribe_operation.activate(subsreq).result(TIMEOUT)
            subscribe_operation.get_response().result(TIMEOUT)
        except Exception as e:
            logging.error(
                "Exception occured during fetching the configuration updates: {}".format(e)
            )
            exit(1)


class FeedbackLoopConfigUpdateHandler(
    awsiot.greengrasscoreipc.client.SubscribeToConfigurationUpdateStreamHandler
):
    r"""
    Custom handle of the subscribed configuration events(steam,error and close).
    Notifies feedbackloop_update_condition on a stream event.
    """

    def on_stream_event(self, event: ConfigurationUpdateEvents) -> None:
        logging.info(event.configuration_update_event)
        with feedbackloop_update_condition:
            feedbackloop_update_condition.notify()

    def on_stream_error(self, error: Exception) -> bool:
        logging.error("Error in config update subscriber - {0}".format(error))
        return False

    def on_stream_closed(self) -> None:
        logging.info("Config update subscription stream was closed")





# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
import logging
import heapq
import math
import random
from abc import ABCMeta, abstractmethod
from numbers import Number

FLOAT_COMPARISON_TOLERANCE = 1e-08

STRATEGY_RANDOM_SAMPLING = "RANDOM_SAMPLING"
STRATEGY_LEAST_CONFIDENCE = "LEAST_CONFIDENCE"
STRATEGY_MARGIN = "MARGIN"
STRATEGY_ENTROPY = "ENTROPY"
SAMPLING_STRATEGIES = {
    STRATEGY_RANDOM_SAMPLING,
    STRATEGY_LEAST_CONFIDENCE,
    STRATEGY_MARGIN,
    STRATEGY_ENTROPY
}


class GreengrassInvalidSamplingStrategyException(Exception):
    '''
        Raised when there is an attempt to create and invalid
        sampling strategy instance.
    '''
    pass


class SamplingStrategy(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def should_use_sample(self, model_prediction):
        pass


class SamplingPercentage(object):
    def __init__(self, value):
        # Check existence using number because threshold can be 0 which
        # evaluates to False
        if not isinstance(value, Number):
            raise GreengrassInvalidSamplingStrategyException(
                "Failed to read percentage value. Requires value to be configured and to be a number."
            )

        if not 0.0 <= value <= 1.0:
            raise GreengrassInvalidSamplingStrategyException(
                "Invalid percentage value: {}. Value must be 0.0 <= percentage <= 1.0".format(str(value))
            )

        self._value = value

    @property
    def value(self):
        return self._value


class RandomSamplingStrategy(SamplingStrategy):
    '''
        SamplingStrategy that randomly selects samples
        to be labeled based on rate. Note: This
        strategy disregards any model prediction that is
        supplied.
    '''

    def __init__(self, rate, seed=None):
        try:
            rate_percentage = SamplingPercentage(rate)
        except Exception as e:
            logging.warn(
                "Failed to create RandomSamplingStrategy because of invalid rate. Failed with exception type: {}, error: {}".format(
                    type(e).__name__, e))
            raise e

        self._rate = rate_percentage.value

        # Used for testing
        if seed:
            random.seed(seed)

    def should_use_sample(self, model_prediction):
        '''
            Returns whether a sample should be used based on
            rate alone. The model_prediction parameter is
            disregarded.
        '''
        random_value = random.random()
        should_use_sample = random_value < self._rate
        logging.info(
            "RandomSamplingStrategy returning {} for should use sample. Should use sample if random generated value ({}) is less than sampling rate ({}).".format(
                str(should_use_sample), str(random_value), str(self._rate)
            ))
        return should_use_sample


class LeastConfidenceSamplingStrategy(SamplingStrategy):
    '''
        SamplingStrategy that selects samples whose max
        confidence falls below some threshold.

        Ex. For model prediction: [.2, .2, .4, .2] and
            threshold: .6

        Max confidence = .4

        Use sample if max confidence (.4) <= threshold (.6)

    '''

    def __init__(self, threshold):
        try:
            threshold_percentage = SamplingPercentage(threshold)
        except Exception as e:
            logging.warn(
                "Failed to create LeastConfidenceSamplingStrategy because of invalid threshold. Failed with exception type: {}, error: {}".format(
                    type(e).__name__, e))
            raise e

        self._threshold = threshold_percentage.value

    def should_use_sample(self, model_prediction):
        '''
            Returns whether a sample should be used based on
            threshold and the model_prediction's highest
            confidence probability.
        '''
        max_confidence = max(model_prediction)
        should_use_sample = max_confidence < self._threshold
        logging.info(
            "LeastConfidenceSamplingStrategy returning {} for should use sample. Should use sample if the highest prediction confidence probability ({}) is less than threshold ({}).".format(
                str(should_use_sample), str(max_confidence), str(self._threshold)
            ))
        return should_use_sample


class MarginSamplingStrategy(SamplingStrategy):
    '''
        SamplingStrategy that selects samples whose top two
        confidence probabiltities fall within a some threshold.

        Ex. For model prediction: [.3, .35, .34, .01] and
            threshold: .02:

        Top two confidence probabilities = [.35, .34]
        Margin = .01 = .35 - .34

        Use sample if margin (.01) <= threshold (.02)
    '''

    def __init__(self, threshold):
        try:
            threshold_percentage = SamplingPercentage(threshold)
        except Exception as e:
            logging.warn(
                "Failed to create MarginSamplingStrategy because of invalid threshold. Failed with exception type: {}, error: {}".format(
                    type(e).__name__, e))
            raise e

        self._threshold = threshold_percentage.value

    def should_use_sample(self, model_prediction):
        '''
           Returns whether a sample should be used based on
           threshold and the model prediction's top two
           highest confidence probabilities.
        '''
        top_two_highest_confidence = heapq.nlargest(2, model_prediction)
        highest_confidence = top_two_highest_confidence[0]
        second_highest_confidence = top_two_highest_confidence[1]
        margin = abs(highest_confidence - second_highest_confidence)
        # Using less than (as opposed to <=) because we are comparing
        # floats.
        should_use_sample = margin < self._threshold
        logging.info(
            "MarginSamplingStrategy returning {} for should use sample. Should use sample if the prediction's top two highest confidence probabilities ({}, {}) have a margin ({}) within threshold ({}).".format(
                str(should_use_sample), str(highest_confidence), str(second_highest_confidence), str(margin),
                str(self._threshold)
            ))
        return should_use_sample


class EntropySamplingStrategy(SamplingStrategy):
    '''
        SamplingStrategy that selects samples whose entropy
        is above some threshold.

        Ex. For model prediction: [.5, .25, .25] and
            threshold: 0.75:

        Entropy for prediction = 1.03972

        Use sample if entropy (1.03972) > threshold (0.75)
    '''

    def __init__(self, threshold):
        if not threshold:
            raise GreengrassInvalidSamplingStrategyException(
                "Failed to create EntropySamplingStrategy. Sampling strategy threshold must be defined.")

        if not isinstance(threshold, Number):
            raise GreengrassInvalidSamplingStrategyException(
                "Failed to create EntropySamplingStrategy. Sampling strategy threshold must be a number.")

        self._threshold = threshold

    @staticmethod
    def _calculate_entropy(model_prediction):
        def _single_class_entropy(class_probability):
            if class_probability > 0:
                return class_probability * math.log(class_probability)
            elif class_probability == 0:
                return 0
            else:
                # Scipy returns -inf when negative probabilies are supplied. We
                # throw here to let customers know that entropy can't be calculated.
                raise ValueError(
                    "Could not calculate entropy. Found negative class probability: {}".format(class_probability))

        # Normalize array to 1.0 first - this is how scipy does
        # this too. See:
        # https://github.com/scipy/scipy/blob/v1.2.1/scipy/stats/_distn_infrastructure.py#L2515-L2556
        model_prediction_probability_sum = sum(model_prediction)
        if abs(1.0 - model_prediction_probability_sum) < FLOAT_COMPARISON_TOLERANCE:
            normalized = model_prediction
        elif model_prediction_probability_sum == 0.0:
            raise ValueError("Could not calculate entropy. Sum of model predictions must not be zero.")
        else:
            normalized = [probability_of_class / sum(model_prediction) for probability_of_class in model_prediction]

        entropy = -sum(_single_class_entropy(class_probability) for class_probability in normalized)
        return entropy

    def should_use_sample(self, model_prediction):
        '''
            Returns whether a sample should be used based on
            threshold and the model prediction's normalized
            entropy. Returns true if the calculated entropy is
            greater than the threshold.
        '''
        entropy = EntropySamplingStrategy._calculate_entropy(model_prediction)
        should_use_sample = entropy > self._threshold
        logging.info(
            "EntropySamplingStrategy returning {} for should use sample. Should use sample if entropy ({}) is greater than threshold ({}).".format(
                str(should_use_sample), str(entropy), str(self._threshold)
            ))
        return should_use_sample


class SamplingStrategyFactory(object):
    '''
        Responsible for creating instances of SamplingStrategy.
    '''

    @staticmethod
    def create_instance(strategy_name, rate=None, threshold=None):
        '''
            Creates an instance of a sampling strategy that
            implements SamplingStrategy. Throws when strategy
            name is not one defined in SAMPLING_STRATEGIES.
        '''
        if strategy_name == STRATEGY_RANDOM_SAMPLING:
            return RandomSamplingStrategy(rate)
        elif strategy_name == STRATEGY_LEAST_CONFIDENCE:
            return LeastConfidenceSamplingStrategy(threshold)
        elif strategy_name == STRATEGY_MARGIN:
            return MarginSamplingStrategy(threshold)
        elif strategy_name == STRATEGY_ENTROPY:
            return EntropySamplingStrategy(threshold)
        else:
            raise GreengrassInvalidSamplingStrategyException(
                "Failed to create SamplingStrategy instance. Found invalid sampling strategy: {}. Sampling strategy must be one of: {}".format(
                    str(strategy_name), str(SAMPLING_STRATEGIES)
                ))



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
import logging
import sys

COUNTER_UPLOAD_OBJECT_SUCCESS = "SampleUploadSuccess"
COUNTER_UPLOAD_OBJECT_FAILURE = "SampleUploadFailure"
BYTES_SAMPLE_SIZE = "SampleSize"

RESPONSE_KEY_S3_BUCKET = "bucket"
RESPONSE_KEY_S3_KEY = "key"

class ModelInputUploader(object):
    '''
        Responsible for uploading the input of ML models
        to S3. Ex. image, sound clip, etc
    '''
    def __init__(self, s3_client, cw_metrics_publisher):
        self._client = s3_client
        self._metrics_publisher = cw_metrics_publisher

    def upload(self, s3_location, model_input, content_type, metadata):
        # Publish size of payload on every request. This may provide
        # a useful hint in case of upload failures.
        payload = open(model_input, 'rb').read()

        self._metrics_publisher.publish_bytes(BYTES_SAMPLE_SIZE, sys.getsizeof(payload))
        try:
            response = self._client.put_object(
                Body=payload,
                Bucket=s3_location.bucket,
                ContentType=content_type,
                Key=s3_location.key.key,
                Metadata=metadata if metadata else {}
            )

            # Add extra fields to the response
            response[RESPONSE_KEY_S3_BUCKET] = s3_location.bucket
            response[RESPONSE_KEY_S3_KEY] = s3_location.key.key

            logging.info("Successfully uploaded model input to S3 location: {}, with response: {}".format(str(s3_location.uri), str(response)))
            self._metrics_publisher.publish_counter(COUNTER_UPLOAD_OBJECT_SUCCESS, 1)
            return response
        except Exception as e:
            logging.warn("Failed to upload object to S3 location: {}, with exception type: {}, error: {}.".format(str(s3_location.uri), type(e).__name__, e))
            self._metrics_publisher.publish_counter(COUNTER_UPLOAD_OBJECT_FAILURE, 1)
            raise e


# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
import logging
from threading import Thread

try:
    import queue
except ImportError:
    import Queue as queue

import awsiot.greengrasscoreipc

from feedbackloop.stream_manager import (
    ExportDefinition,
    ResourceNotFoundException,
    MessageStreamDefinition,
    Persistence,
    S3ExportTaskExecutorConfig,
    Status,
    StatusConfig,
    StatusLevel,
    StatusMessage,
    StrategyOnFull,
    StreamManagerClient,
)

from feedbackloop.service.IPCUtils import IPCUtils, feedbackloop_update_condition
from feedbackloop.model.feedback_input import FeedbackInput, PARAM_CONFIG_ID, PARAM_MODEL_PREDICTION, \
                                              PARAM_METADATA, PARAM_STATUS_TOPIC, PARAM_PREDICTION_TOPIC, PARAM_FILENAME
from feedbackloop.model.component_config import (
    ComponentConfig, PARAM_FEEDBACK_CONFIG_MAP, PARAM_MAX_REQUESTS)
from feedbackloop.request_processor_sm import RequestProcessor
from feedbackloop.service.cloudwatch_metrics_publisher import CWComponentMetricsPublisher
from feedbackloop.service.model_input_uploader_sm import ModelInputUploaderSM
from feedbackloop.service.model_prediction_publisher import ModelPredictionPublisher, DEFAULT_PREDICTION_TOPIC
from feedbackloop.service.status_publisher import StatusPublisher, StatusMessageError, DEFAULT_STATUS_TOPIC

COUNTER_INVALID_COMPONENT_CONFIG = "ErrorInvalidComponentConfig"
COUNTER_INVALID_INPUT = "ErrorInvalidInput"
COUNTER_REQUEST_THROTTLED = "ErrorRequestThrottled"

ipc_client = awsiot.greengrasscoreipc.connect()
ipc_utils = IPCUtils(ipc_client)
model_prediction_publisher = ModelPredictionPublisher(ipc_client)
metrics_publisher = CWComponentMetricsPublisher(ipc_client)
status_publisher = StatusPublisher(ipc_client)

s3_client = StreamManagerClient()

model_input_uploader = ModelInputUploaderSM(s3_client, metrics_publisher)


def create_stream(stream_name=None, parameters=None):
    '''
        Create stream with name stream_name with the given parameters
        If stream already exists, warn user and return without doing anything.
    '''
    if not isinstance(stream_name, str):
        raise ValueError("stream_name argument must be of type string.")

    if not isinstance(parameters, dict):
        raise ValueError("parameters argument must be of type dict.")

    streams = s3_client.list_streams()
    if stream_name in streams:
        logging.warn("Stream with name [{}] already exists".format(stream_name))
        return

    status_stream_name = stream_name + "Status"

    exports = ExportDefinition(
        s3_task_executor=[
            S3ExportTaskExecutorConfig(
                identifier="S3TaskExecutor" + stream_name,  # Required
                # Optional. Add an export status stream to add statuses for all S3 upload tasks.
                status_config=StatusConfig(
                    status_level=StatusLevel.INFO,  # Default is INFO level statuses.
                    # Status Stream should be created before specifying in S3 Export Config.
                    status_stream_name=status_stream_name,
                ),
            )
        ]
    )

    DEFAULT_MAX_SIZE = 268435456
    DEFAULT_SEGMENT_SIZE = 16777216

    max_size = parameters.get("max_size") if parameters.get("max_size") else DEFAULT_MAX_SIZE
    stream_segment_size = parameters.get("stream_segment_size") if parameters.get("stream_segment_size") else DEFAULT_SEGMENT_SIZE
    time_to_live_millis = parameters.get("time_to_live_millis")

    strategy_on_full = parameters.get("strategy_on_full")
    if strategy_on_full == 0:
        strategy_on_full = StrategyOnFull.RejectNewData
    else:
        #default is OverwriteOldestData
        strategy_on_full = StrategyOnFull.OverwriteOldestData

    persistence = parameters.get("persistence")
    if persistence == 1:
        persistence = Persistence.Memory
    else:
        #default is File
        persistence = Persistence.File
    flush_on_write = parameters.get("flush_on_write")

    # Create the Status Stream with default parameters
    s3_client.create_message_stream(
        MessageStreamDefinition(name=status_stream_name, strategy_on_full=StrategyOnFull.OverwriteOldestData)
    )

    # Create the message stream with the S3 Export definition and given parameters.
    s3_client.create_message_stream(
        MessageStreamDefinition(
            name=stream_name,
            max_size=max_size,
            stream_segment_size=stream_segment_size,
            time_to_live_millis=time_to_live_millis,
            strategy_on_full=strategy_on_full,
            persistence=persistence,
            flush_on_write=flush_on_write,
            export_definition=exports
        )
    )


def update_stream(stream_name, parameters):
    '''
        Update stream with name stream_name with the given parameters
        If no such stream exists, raise ValueError
    '''
    if not isinstance(stream_name, str):
        raise ValueError("stream_name argument must be of type string.")

    if not isinstance(parameters, dict):
        raise ValueError("parameters argument must be of type dict.")

    streams = s3_client.list_streams()
    if stream_name in streams:
        status_stream_name = stream_name + "Status"

        exports = ExportDefinition(
            s3_task_executor=[
                S3ExportTaskExecutorConfig(
                    identifier="S3TaskExecutor" + stream_name,  # Required
                    # Optional. Add an export status stream to add statuses for all S3 upload tasks.
                    status_config=StatusConfig(
                        status_level=StatusLevel.INFO,  # Default is INFO level statuses.
                        # Status Stream should be created before specifying in S3 Export Config.
                        status_stream_name=status_stream_name,
                    ),
                )
            ]
        )

        DEFAULT_MAX_SIZE = 268435456
        DEFAULT_SEGMENT_SIZE = 16777216

        max_size = parameters.get("max_size") if parameters.get("max_size") else DEFAULT_MAX_SIZE
        stream_segment_size = parameters.get("stream_segment_size") if parameters.get(
            "stream_segment_size") else DEFAULT_SEGMENT_SIZE
        time_to_live_millis = parameters.get("time_to_live_millis")

        strategy_on_full = parameters.get("strategy_on_full")
        if strategy_on_full == 0:
            strategy_on_full = StrategyOnFull.RejectNewData
        else:
            # default is OverwriteOldestData
            strategy_on_full = StrategyOnFull.OverwriteOldestData

        persistence = parameters.get("persistence")
        if persistence == 1:
            persistence = Persistence.Memory
        else:
            # default is File
            persistence = Persistence.File
        flush_on_write = parameters.get("flush_on_write")

        # Create the message stream with the S3 Export definition and given parameters.
        s3_client.update_message_stream(
            MessageStreamDefinition(
                name=stream_name,
                max_size=max_size,
                stream_segment_size=stream_segment_size,
                time_to_live_millis=time_to_live_millis,
                strategy_on_full=strategy_on_full,
                persistence=persistence,
                flush_on_write=flush_on_write,
                export_definition=exports
            )
        )
    else:
        logging.warn("Stream with name [{}] does not exist exists".format(stream_name))


def delete_stream(stream_name):
    '''
        Deletes a stream with name stream_name.
        If no stream with name stream_name, pass.
    '''
    if not isinstance(stream_name, str):
        raise ValueError("stream_name argument must be of type string.")

    # Try deleting the stream (if it exists)
    try:
        s3_client.delete_message_stream(stream_name=stream_name)
    except ResourceNotFoundException:
        pass

    status_stream_name = stream_name + "Status"
    try:
        s3_client.delete_message_stream(stream_name=status_stream_name)
    except ResourceNotFoundException:
        pass


def modify_configuration(new_configuration):
    try:
        # This try block is for catching user configuration errors
        new_component_param_feedback_config = new_configuration.get(PARAM_FEEDBACK_CONFIG_MAP)
        new_component_param_max_requests = new_configuration.get(PARAM_MAX_REQUESTS)
        new_component_config = ComponentConfig(
            int(new_component_param_max_requests if new_component_param_feedback_config else 0),
            new_component_param_feedback_config
        )
        return new_component_config
    except Exception as e:
        metrics_publisher.publish_counter(COUNTER_INVALID_COMPONENT_CONFIG, 1)
        error_str = "Failed to initialize. Exception raised while parsing component configuration. Stopping component execution. Failed with: {}".format(
            e)
        logging.error(error_str)
        raise e


# Listen to see if feedback configuration map changes, if so, update the current component configuration
def wait_for_config_changes():
    with feedbackloop_update_condition:
        feedbackloop_update_condition.wait()
        global component_config
        component_config = modify_configuration(ipc_utils.get_feedbackloop_configuration())
    wait_for_config_changes()


# initialize component config
component_config = modify_configuration(ipc_utils.get_feedbackloop_configuration())

try:
    # This try block is for catching unexpected errors
    request_queue = queue.Queue(component_config.max_requests)
    request_processor = RequestProcessor(
        request_queue=request_queue,
        metrics_publisher=metrics_publisher,
        status_publisher=status_publisher,
        model_input_uploader=model_input_uploader,
        model_prediction_publisher=model_prediction_publisher
    )

    # request_processor runs on a separate thread and listens for
    # events published to the request_queue.
    request_processor.start()
except Exception as e:
    error_str = "Failed to initialize. Unexpected exception raised:{}".format(e)
    logging.error(error_str)
    raise e

logging.info("Successfully initialized the feedback component.")
# Subscribe to the subsequent configuration changes
ipc_utils.get_feedbackloop_config_updates()

Thread(
    target=wait_for_config_changes,
    args=(),
).start()


def handle_error(exception, error_str, counter, topic):
    metrics_publisher.publish_counter(counter, 1)
    logging.warn(error_str)
    error_message = StatusMessageError(exception=exception, message=error_str)
    status_publisher.publish(error_message, topic)


def function_handler(model_input, handler_data, stream_name):
    # Validate input and return to client immediately if something is
    # wrong.
    status_topic = handler_data.get(PARAM_STATUS_TOPIC)

    try:
        feedback_input = FeedbackInput(model_input, handler_data)
        if feedback_input.config_id not in component_config.feedback_config_map:
            raise ValueError("Feedback configuration map does not contain config_id: {}".format(feedback_input.config_id))
        feedback_config = component_config.feedback_config_map[feedback_input.config_id]
    except Exception as e:
        error_str = "Exception raised while parsing input. Failed with exception type: {}, error: {}".format(type(e).__name__, e)
        handle_error(exception=e, error_str=error_str, counter=COUNTER_INVALID_INPUT, topic=status_topic)
        raise e

    try:
        logging.info("Enqueing request to be processed.")
        request_queue.put_nowait((feedback_input, feedback_config, stream_name))
    except queue.Full as e:
        error_str = "Request limit has been reached (max request:{}). Dropping request.".format(component_config.max_requests)
        handle_error(exception=e, error_str=error_str, counter=COUNTER_REQUEST_THROTTLED, topic=status_topic)
        raise e

    logging.info("Request has successfully be enqueued.")
    return


def publish(model_input,
            model_prediction,
            config_id,
            stream_name,
            filename=None,
            prediction_topic=None,
            status_topic=None,
            metadata=None):
    """
    Customer facing function that is responsible for packaging feedback input data
    (e.g. config id, model prediction, etc.) and model input (ex: inference images)
    together in format readable by the feedback loop handler
    """
    prediction_topic = prediction_topic if prediction_topic else DEFAULT_PREDICTION_TOPIC
    status_topic = status_topic if status_topic else DEFAULT_STATUS_TOPIC

    feedback_input_data = {
        PARAM_CONFIG_ID: config_id,
        PARAM_MODEL_PREDICTION: model_prediction,
        PARAM_METADATA: metadata,
        PARAM_FILENAME: filename,
        PARAM_PREDICTION_TOPIC: prediction_topic,
        PARAM_STATUS_TOPIC: status_topic
    }

    # pass model input and handler_data onto data handler
    function_handler(
        model_input=model_input,
        handler_data=feedback_input_data,
        stream_name=stream_name
    )


# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
import logging
import sys
import time

from feedbackloop.stream_manager import (
    ReadMessagesOptions,
    S3ExportTaskDefinition,
    Status,
    StatusMessage,
    StreamManagerException,
)
from feedbackloop.stream_manager.util import Util

COUNTER_UPLOAD_OBJECT_SUCCESS = "SampleUploadSuccess"
COUNTER_UPLOAD_OBJECT_FAILURE = "SampleUploadFailure"
BYTES_SAMPLE_SIZE = "SampleSize"

RESPONSE_KEY_S3_BUCKET = "bucket"
RESPONSE_KEY_S3_KEY = "key"

FILE_URI_PREFIX = "file:"


class ModelInputUploaderSM(object):
    '''
        Responsible for uploading the input of ML models
        to S3. Ex. image, sound clip, etc
    '''
    def __init__(self, sm_client, cw_metrics_publisher):
        self._client = sm_client
        self._metrics_publisher = cw_metrics_publisher

    def upload(self, s3_location, model_input, content_type, stream_name, metadata):
        # Publish size of payload on every request. This may provide
        # a useful hint in case of upload failures.
        payload = open(model_input, 'rb').read()

        self._metrics_publisher.publish_bytes(BYTES_SAMPLE_SIZE, sys.getsizeof(payload))

        # Append a S3 Task definition and print the sequence number
        try:
            file_uri = FILE_URI_PREFIX + model_input
            s3_export_task_definition = S3ExportTaskDefinition(input_url=file_uri,
                                                               bucket=s3_location.bucket,
                                                               key=s3_location.key.key,
                                                               user_metadata=metadata if metadata else {})
            logging.info(
                "Successfully appended S3 Task Definition to stream with sequence number %d",
                self._client.append_message(stream_name, Util.validate_and_serialize_to_json_bytes(s3_export_task_definition)),
            )
        except Exception as e:
            logging.info("Failed to append S3 Task Definition")

        # Read the statuses from the export status stream
        stop_checking = False
        next_seq = 0
        response = {
            "bucket": s3_location.bucket,
            "key": s3_location.key.key
        }

        while not stop_checking:
            try:
                status_stream_name = stream_name + "Status"
                messages_list = self._client.read_messages(
                    status_stream_name,
                    ReadMessagesOptions(
                        desired_start_sequence_number=next_seq, min_message_count=1, read_timeout_millis=1000
                    ),
                )
                for message in messages_list:
                    # Deserialize the status message first.
                    status_message = Util.deserialize_json_bytes_to_obj(message.payload, StatusMessage)

                    # Check the status of the status message. If the status is "Success",
                    # the file was successfully uploaded to S3.
                    # If the status was either "Failure" or "Cancelled", the server was unable to upload the file to S3.
                    # If the status was "InProgress", this indicates that the server has started uploading the S3 task.
                    if status_message.status == Status.Success:
                        # Add extra fields to the response
                        logging.info("Successfully uploaded model input {} to S3 location: {}.".format(model_input, str(s3_location.uri)))
                        self._metrics_publisher.publish_counter(COUNTER_UPLOAD_OBJECT_SUCCESS, 1)
                        stop_checking = True
                    elif status_message.status == Status.InProgress:
                        logging.info("File upload is in Progress.")
                        next_seq = message.sequence_number + 1
                    elif status_message.status == Status.Failure or status_message.status == Status.Canceled:
                        logging.warn("Failed to upload object to S3 location: {}, with exception type: {}, error: {}.".format(str(s3_location.uri), type(e).__name__, e))
                        self._metrics_publisher.publish_counter(COUNTER_UPLOAD_OBJECT_FAILURE, 1)
                        stop_checking = True
                if not stop_checking:
                    time.sleep(5)
            except StreamManagerException:
                logging.exception("Exception while running")
                time.sleep(5)

        return response





# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.

import re
import six
from feedbackloop.sampling.sampling_strategy import SamplingStrategyFactory

PARAM_S3_BUCKET = "s3-bucket-name"
PARAM_S3_PREFIX = "s3-prefix"
PARAM_FILE_EXTENSION = "file-ext"
PARAM_CONTENT_TYPE = "content-type"
PARAM_SAMPLING_STRATEGY = "sampling-strategy"
PARAM_SAMPLING_STRATEGY_NAME = "strategy-name"
PARAM_SAMPLING_RATE = "rate"
PARAM_SAMPLING_THRESHOLD = "threshold"

PATTERN_STRING_S3_BUCKET = "^[a-z0-9\.\-]{3,63}$"
PATTERN_S3_BUCKET = re.compile(PATTERN_STRING_S3_BUCKET)

COMMON_CONSTRUCTOR_ERR_MSG = "Failed to construct feedback configuration entry."


class FeedbackConfig (object):
    '''
    Represents a particular feedback configuration which contains the following:

    s3_bucket (string): The name of the s3 bucket that data is uploaded to.
    content_type (string): The content type of the data that is uploaded (ex: image/jpg).
    s3_prefix (string): (Optional) s3 prefix for uploaded data, if necessary.
    file_ext (string): The file extension for the files that are uploaded (ex: jpg, txt).
    sampling_strategy (SamplingStrategy): (Optional) The particular sampling strategy to use when uploading data.
                                          See sampling/sampling_strategy.py for more information/details
    '''

    def __init__(self, feedback_config_entry_json):
        # Required parameters
        self._s3_bucket = feedback_config_entry_json[PARAM_S3_BUCKET]
        # Do basic regex checking for invalid characters and length
        if not PATTERN_S3_BUCKET.match(self._s3_bucket):
            raise ValueError("{} Found invalid S3 bucket name: {}".format(
                COMMON_CONSTRUCTOR_ERR_MSG, str(self._s3_bucket))
            )

        self._content_type = feedback_config_entry_json[PARAM_CONTENT_TYPE]
        if not isinstance(self._content_type, six.string_types):
            raise ValueError("{} Content type argument must be of type string.".format(COMMON_CONSTRUCTOR_ERR_MSG))

        # Optional parameters
        self._s3_prefix = feedback_config_entry_json.get(PARAM_S3_PREFIX)
        if self._s3_prefix:
            if not isinstance(self._s3_prefix, six.string_types):
                raise ValueError("{} S3 prefix argument must be of type string.".format(COMMON_CONSTRUCTOR_ERR_MSG))
        else:
            self._s3_prefix = ""

        self._file_ext = feedback_config_entry_json.get(PARAM_FILE_EXTENSION)
        if self._file_ext:
            if not isinstance(self._file_ext, six.string_types):
                raise ValueError("{} File extension argument must be of type string.".format(COMMON_CONSTRUCTOR_ERR_MSG))
        else:
            self._file_ext = ""

        sampling_strategy_config = feedback_config_entry_json.get(PARAM_SAMPLING_STRATEGY)
        if sampling_strategy_config:
            self._sampling_strategy = SamplingStrategyFactory.create_instance(
                strategy_name=sampling_strategy_config[PARAM_SAMPLING_STRATEGY_NAME],
                rate=sampling_strategy_config.get(PARAM_SAMPLING_RATE),
                threshold=sampling_strategy_config.get(PARAM_SAMPLING_THRESHOLD)
            )
        else:
            self._sampling_strategy = None

    @property
    def s3_bucket(self):
        return self._s3_bucket

    @property
    def s3_prefix(self):
        return self._s3_prefix

    @property
    def file_ext(self):
        return self._file_ext

    @property
    def content_type(self):
        return self._content_type

    @property
    def sampling_strategy(self):
        return self._sampling_strategy



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.


class Filename(object):
    def __init__(self, filename, file_extension):
        if file_extension:
            stripped_file_extension = file_extension.strip('.')
            self._filename = "{}.{}".format(filename, stripped_file_extension)
        else:
            self._filename = filename

    @property
    def filename(self):
        return self._filename



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.

import json

from feedbackloop.model.feedback_config import FeedbackConfig
from feedbackloop.model.config_id import ConfigId

PARAM_FEEDBACK_CONFIG_MAP = "FEEDBACK_CONFIGURATION_MAP"
PARAM_MAX_REQUESTS = "REQUEST_LIMIT"

ERROR_PREFIX_CONSTRUCT_MAP = "Failed to build feedback configuration map."


class ComponentConfig (object):
    '''
    Stores the Feedback Loop Component Configuration.
    This consists of:
        1. max_requests (int): Represents the maximum number of requests the Feedback Loop
                               request processor queue can handle at one time.

        2. feedback_config_map (Map<config-id, FeedbackConfig>): User defined configuration map

    '''
    def __init__(self, max_requests, feedback_config_json_str):
        if not isinstance(max_requests, int):
            raise ValueError("Failed to parse component configuration. Max requests must be of type 'int'.")
        self._max_requests = max_requests

        # load json with custom hook to raise on
        # duplicate config ids.
        feedback_config_map_json = json.loads(
            feedback_config_json_str,
            object_pairs_hook=ComponentConfig._load_json_raise_on_duplicate_keys
        )
        if not isinstance(feedback_config_map_json, dict):
            raise ValueError("Failed to parse component configuration. Feedback configuration JSON must be a valid dict.")

        self._feedback_config_map = self._build_feedback_config_map(feedback_config_map_json)

    @staticmethod
    def _load_json_raise_on_duplicate_keys(key_values):
        json_map = {}
        for key, value in key_values:
            if key in json_map:
                raise ValueError("{} Found duplicate keys: {}. Keys must be unique.".format(
                    ERROR_PREFIX_CONSTRUCT_MAP, str(key))
                )
            else:
                json_map[key] = value
        return json_map

    @staticmethod
    def _build_feedback_config_map(feedback_config_map_json):
        '''
            Marshals raw json map into Map<config-id, FeedbackConfig>
        '''
        feedback_config_map = {}
        for config_id_str, config_entry_json in feedback_config_map_json.items():
            config_id = ConfigId(config_id_str)
            feedback_config = FeedbackConfig(config_entry_json)
            feedback_config_map[config_id.id] = feedback_config
        if not feedback_config_map:
            raise ValueError("{} Failed to build feedback configuration list. No configurations specified.".format(
                ERROR_PREFIX_CONSTRUCT_MAP
            ))

        return feedback_config_map

    @property
    def max_requests(self):
        return self._max_requests

    @property
    def feedback_config_map(self):
        '''
            Returns Map<Feedback_Config_Id, FeedbackConfig>
        '''
        return self._feedback_config_map



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
import json
import logging
import numbers
import six

from awsiot.greengrasscoreipc.model import (
    PublishToTopicRequest,
    PublishMessage,
    BinaryMessage
)


CLOUD_WATCH_NAME_SPACE = "GreengrassFeedbackComponent"
DIMENSION_FEEDBACK_CONFIG_ID = "FeedbackConfigId"
TOPIC_CLOUD_WATCH_COMPONENT = "cloudwatch/metric/put"


class CWComponentMetricsPublisher (object):

    def __init__(self, gg_client):
        self._iot_client = gg_client

    def _publish_message(self, payload):
        '''
        Responsible for publishing cloudwatch metrics to cloudwatch component through IPC.
        '''
        topic = TOPIC_CLOUD_WATCH_COMPONENT
        TIMEOUT = 10
        request = PublishToTopicRequest()
        request.topic = topic

        publish_message = PublishMessage()
        publish_message.binary_message = BinaryMessage()
        publish_message.binary_message.message = bytes(json.dumps(payload), "utf-8")
        request.publish_message = publish_message

        operation = self._iot_client.new_publish_to_topic()
        operation.activate(request)
        future = operation.get_response()
        future.result(TIMEOUT)

    def _publish_value(self, metric_name, value, unit, config_id):
        logging.info("Publishing to Cloudwatch component with metric_name: {}, value: {}, unit: {}, config_id: {}.".format(str(metric_name), str(value), str(unit), str(config_id)))

        dimensions = None
        if config_id:
            dimensions = []
            config_dimension = MetricDimension(DIMENSION_FEEDBACK_CONFIG_ID, config_id)
            dimensions.append(config_dimension)

        metric_data = MetricData(
            metric_name=metric_name,
            value=value,
            dimensions=dimensions,
            unit=unit
        )

        request_message = CWComponentMessage(
            namespace=CLOUD_WATCH_NAME_SPACE,
            metric_data=metric_data
        )

        self._publish_message(request_message.json)

    def publish_counter(self, metric_name, value, config_id=None):
        self._publish_value(metric_name=metric_name, value=value, unit="Count", config_id=config_id)

    def publish_bytes(self, metric_name, value, config_id=None):
        self._publish_value(metric_name=metric_name, value=value, unit="Bytes", config_id=config_id)


class CWComponentMessage (object):
    EXCEPTION_PREFIX = "Failed to construct CWComponentMessage."

    def __init__(self, namespace, metric_data):
        if not isinstance(namespace, six.string_types):
            raise ValueError("{} Requires namespace to be of type string".format(
                self.EXCEPTION_PREFIX
            ))
        self._namespace = namespace

        if not isinstance(metric_data, MetricData):
            raise ValueError("{} Requires metric_data to be of type MetricData".format(
                self.EXCEPTION_PREFIX
            ))
        self._metric_data = metric_data

    @property
    def json(self):
        return {
            "request": {
                "namespace": self._namespace,
                "metricData": self._metric_data.json
            }
        }


class MetricData (object):
    EXCEPTION_PREFIX = "Failed to construct MetricData."

    def __init__(self, metric_name, value, dimensions=None, timestamp=None, unit=None):
        if not isinstance(metric_name, six.string_types):
            raise ValueError("{} Requires metric_name to be of type string".format(
                self.EXCEPTION_PREFIX
            ))
        self._metric_name = metric_name

        if not isinstance(value, numbers.Number):
            raise ValueError("{} Requires value to be of type number".format(
                self.EXCEPTION_PREFIX
            ))
        self._value = value

        if dimensions:
            if not isinstance(dimensions, list):
                raise ValueError("{} Requires dimensions to be a list".format(
                    self.EXCEPTION_PREFIX
                ))
            for dimension in dimensions:
                if not isinstance(dimension, MetricDimension):
                    raise ValueError("{} Requires dimension to be MetricDimension".format(
                        self.EXCEPTION_PREFIX
                    ))
        self._dimensions = dimensions

        if timestamp:
            if not isinstance(timestamp, float):
                raise ValueError("{} Requires timestamp to be of type float".format(
                    self.EXCEPTION_PREFIX
                ))
        self._timestamp = timestamp

        if unit:
            if not isinstance(unit, six.string_types):
                raise ValueError("{} Requires unit to be of type string".format(
                    self.EXCEPTION_PREFIX
                ))
        self._unit = unit

    @property
    def json(self):
        metric_data = {
            "metricName": self._metric_name,
            "value": self._value,
        }

        if self._dimensions:
            dimensions_to_add = []
            for dimension in self._dimensions:
                dimensions_to_add.append(dimension.json)
            metric_data["dimensions"] = dimensions_to_add

        if self._timestamp:
            metric_data["timestamp"] = self._timestamp

        if self._unit:
            metric_data["unit"] = self._unit

        return metric_data


class MetricDimension (object):
    def __init__(self, name, value):
        if not isinstance(name, six.string_types):
            raise ValueError("Found invalid metric dimension name. Name must be a string.")

        if not isinstance(value, six.string_types):
            raise ValueError("Found invalid metric dimension value. Value must be a string.")

        self._name = name
        self._value = value

    @property
    def json(self):
        return {
            "name": self._name,
            "value": self._value
        }



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
import json
import logging
import feedbackloop.service.status_messages as StatusResponse
from abc import ABCMeta, abstractmethod

from awsiot.greengrasscoreipc.model import (
    PublishToIoTCoreRequest
)

from feedbackloop.service.IPCUtils import QOS_TYPE, TIMEOUT

FIELD_S3_RESPONSE = "s3_response"
VALUE_SAMPLE_DROPPED = "sample_dropped_by_strategy"

DEFAULT_STATUS_TOPIC = "feedback/message/status"

class StatusMessage(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def payload(self):
        pass


class StatusMessageSuccess(StatusMessage):
    def __init__(self, s3_response_json):
        response_fields = {
            FIELD_S3_RESPONSE : s3_response_json
        }
        self._response = StatusResponse.generate_success_response(**response_fields)

    @property
    def payload(self):
        return json.dumps(self._response)


class StatusMessageSampleDropped(StatusMessage):
    def __init__(self):
        success_message = StatusResponse.generate_success_response()
        success_message[StatusResponse.RESPONSE_FIELD_STATUS] = VALUE_SAMPLE_DROPPED
        self._response = success_message

    @property
    def payload(self):
        return json.dumps(self._response)


class StatusMessageError(StatusMessage):
    def __init__(self, exception, message):
        if not isinstance(exception, Exception):
            raise ValueError("StatusMessageError requires exception to be of type Exception.")
        self._response = StatusResponse.generate_error_response(type(exception).__name__, message)

    @property
    def payload(self):
        return json.dumps(self._response)


class StatusPublisher(object):
    def __init__(self, iot_client):
        self._client = iot_client

    def publish(self, status_message, status_topic):
        if not isinstance(status_message, StatusMessage):
            raise ValueError("status_message must be of type StatusMessage.")

        logging.info("Publishing status message to topic: {}, with message payload: {}".format(
            status_topic, str(status_message.payload)
        ))

        request = PublishToIoTCoreRequest()
        request.topic_name = status_topic
        request.qos = QOS_TYPE

        request.payload = bytes(status_message.payload, "utf-8")

        operation = self._client.new_publish_to_iot_core()
        operation.activate(request)
        future = operation.get_response()
        future.result(TIMEOUT)



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
import re
import six

PATTERN_STRING_CONFIG_ID = "^[a-zA-Z0-9][a-zA-Z0-9-]{1,62}$"
PATTERN_CONFIG_ID = re.compile(PATTERN_STRING_CONFIG_ID)


class ConfigId(object):
    '''
        Responsible for validating and storing a feedback
        configuration id. Valid ids must be:
        1. At least 2 characters
        2. Less than or equal to 63 characters
        3. Lead with a lower/upper case letter or number
        4. Contain lower/upper case letters, numbers, or hyphens
        Examples:
          config-a
          Config-1
          id0
    '''
    def __init__(self, config_id_str):
        if not isinstance(config_id_str, six.string_types) or not PATTERN_CONFIG_ID.match(config_id_str):
            raise ValueError(
                "Failed to create config id. Config id is invalid: {}. Config id must match: {}".format(
                    str(config_id_str), PATTERN_STRING_CONFIG_ID
                )
            )
        self._id = config_id_str

    @property
    def id(self):
        return self._id



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
import boto3
import logging
from threading import Thread

try:
    import queue
except ImportError:
    import Queue as queue

import awsiot.greengrasscoreipc

from feedbackloop.service.IPCUtils import IPCUtils, feedbackloop_update_condition
from feedbackloop.model.feedback_input import FeedbackInput, PARAM_CONFIG_ID, PARAM_MODEL_PREDICTION, \
                                              PARAM_METADATA, PARAM_STATUS_TOPIC, PARAM_PREDICTION_TOPIC, PARAM_FILENAME
from feedbackloop.model.component_config import (
    ComponentConfig, PARAM_FEEDBACK_CONFIG_MAP, PARAM_MAX_REQUESTS)
from feedbackloop.request_processor_s3 import RequestProcessor
from feedbackloop.service.cloudwatch_metrics_publisher import CWComponentMetricsPublisher
from feedbackloop.service.model_input_uploader import ModelInputUploader
from feedbackloop.service.model_prediction_publisher import ModelPredictionPublisher, DEFAULT_PREDICTION_TOPIC
from feedbackloop.service.status_publisher import StatusPublisher, StatusMessageError, DEFAULT_STATUS_TOPIC


COUNTER_INVALID_COMPONENT_CONFIG = "ErrorInvalidComponentConfig"
COUNTER_INVALID_INPUT = "ErrorInvalidInput"
COUNTER_REQUEST_THROTTLED = "ErrorRequestThrottled"


s3_client = boto3.client('s3')
ipc_client = awsiot.greengrasscoreipc.connect()
ipc_utils = IPCUtils(ipc_client)
model_prediction_publisher = ModelPredictionPublisher(ipc_client)
metrics_publisher = CWComponentMetricsPublisher(ipc_client)
status_publisher = StatusPublisher(ipc_client)
model_input_uploader = ModelInputUploader(s3_client, metrics_publisher)


def modify_configuration(new_configuration):
    try:
        # This try block is for catching user configuration errors
        new_component_param_feedback_config = new_configuration.get(PARAM_FEEDBACK_CONFIG_MAP)
        new_component_param_max_requests = new_configuration.get(PARAM_MAX_REQUESTS)
        new_component_config = ComponentConfig(
            int(new_component_param_max_requests if new_component_param_feedback_config else 0),
            new_component_param_feedback_config
        )
        return new_component_config
    except Exception as e:
        metrics_publisher.publish_counter(COUNTER_INVALID_COMPONENT_CONFIG, 1)
        error_str = "Failed to initialize. Exception raised while parsing component configuration. Stopping component execution. Failed with: {}".format(
            e)
        logging.error(error_str)
        raise e


# Listen to see if feedback configuration map changes, if so, update the current component configuration
def wait_for_config_changes():
    with feedbackloop_update_condition:
        feedbackloop_update_condition.wait()
        global component_config
        component_config = modify_configuration(ipc_utils.get_feedbackloop_configuration())
    wait_for_config_changes()


# initialize component config
component_config = modify_configuration(ipc_utils.get_feedbackloop_configuration())

try:
    # This try block is for catching unexpected errors
    request_queue = queue.Queue(component_config.max_requests)
    request_processor = RequestProcessor(
        request_queue=request_queue,
        metrics_publisher=metrics_publisher,
        status_publisher=status_publisher,
        model_input_uploader=model_input_uploader,
        model_prediction_publisher=model_prediction_publisher
    )

    # request_processor runs on a separate thread and listens for
    # events published to the request_queue.
    request_processor.start()
except Exception as e:
    error_str = "Failed to initialize. Unexpected exception raised:{}".format(e)
    logging.error(error_str)
    raise e

logging.info("Successfully initialized the feedback component.")

# Subscribe to the subsequent configuration changes
ipc_utils.get_feedbackloop_config_updates()

Thread(
    target=wait_for_config_changes,
    args=(),
).start()


def handle_error(exception, error_str, counter, topic):
    metrics_publisher.publish_counter(counter, 1)
    logging.warn(error_str)
    error_message = StatusMessageError(exception=exception, message=error_str)
    status_publisher.publish(error_message, topic)


def function_handler(model_input, handler_data):
    # Validate input and return to client immediately if something is
    # wrong.
    status_topic = handler_data.get(PARAM_STATUS_TOPIC)

    try:
        feedback_input = FeedbackInput(model_input, handler_data)
        if feedback_input.config_id not in component_config.feedback_config_map:
            raise ValueError("Feedback configuration map does not contain config_id: {}".format(feedback_input.config_id))
        feedback_config = component_config.feedback_config_map[feedback_input.config_id]
    except Exception as e:
        error_str = "Exception raised while parsing input. Failed with exception type: {}, error: {}".format(type(e).__name__, e)
        handle_error(exception=e, error_str=error_str, counter=COUNTER_INVALID_INPUT, topic=status_topic)
        raise e

    try:
        logging.info("Enqueing request to be processed.")
        request_queue.put_nowait((feedback_input, feedback_config))
    except queue.Full as e:
        error_str = "Request limit has been reached (max request:{}). Dropping request.".format(component_config.max_requests)
        handle_error(exception=e, error_str=error_str, counter=COUNTER_REQUEST_THROTTLED, topic=status_topic)
        raise e

    logging.info("Request has successfully be enqueued.")
    return


def publish(model_input,
            model_prediction,
            config_id,
            filename=None,
            prediction_topic=None,
            status_topic=None,
            metadata=None):
    """
    Customer facing function that is responsible for packaging feedback input data
    (e.g. config id, model prediction, etc.) and model input (ex: inference images)
    together in format readable by the feedback loop handler.
    """
    prediction_topic = prediction_topic if prediction_topic else DEFAULT_PREDICTION_TOPIC
    status_topic = status_topic if status_topic else DEFAULT_STATUS_TOPIC

    feedback_input_data = {
        PARAM_CONFIG_ID: config_id,
        PARAM_MODEL_PREDICTION: model_prediction,
        PARAM_METADATA: metadata,
        PARAM_FILENAME: filename,
        PARAM_PREDICTION_TOPIC: prediction_topic,
        PARAM_STATUS_TOPIC: status_topic
    }

    # pass model input and handler_data onto data handler
    function_handler(
        model_input=model_input,
        handler_data=feedback_input_data
    )


from setuptools import setup, find_packages
import codecs
import os

VERSION = '1.0.2'
DESCRIPTION = 'feedback loop greengrass ML component'
LONG_DESCRIPTION = 'a package that allows greengrass devicdes to upload ML inference data to the cloud'

# Setting up
setup(
    name="feedbackloop",
    version=VERSION,
    author="William Xi",
    author_email="williaxi@amazon.com",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=['boto3', 'awsiotsdk'],
    keywords=['python', 'iot', 'aws', 'machine learning'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.

from feedbackloop.model.config_id import ConfigId
PARAM_CONFIG_ID = "ConfigId"
PARAM_MODEL_PREDICTION = "ModelPrediction"
PARAM_METADATA = "Metadata"
PARAM_CONFIG_ID = "FeedbackConfigID"
PARAM_PREDICTION_TOPIC = "PublishPredictionsOnTopic"
PARAM_STATUS_TOPIC = "PublishStatusOnTopic"
PARAM_FILENAME = "Filename"


class FeedbackInput (object):
    '''
        Responsible for validating and storing IPC requests
        to the feedback component.
    '''

    def __init__(self, model_input, context):
        if PARAM_CONFIG_ID not in context:
            raise ValueError("Failed to construct FeedbackInput. Context is missing param: {}".format(PARAM_CONFIG_ID))
        self._config_id = ConfigId(context[PARAM_CONFIG_ID])

        if PARAM_MODEL_PREDICTION not in context:
            raise ValueError("Failed to construct FeedbackInput. Context is missing param: {}".format(PARAM_MODEL_PREDICTION))
        model_prediction = context[PARAM_MODEL_PREDICTION]
        # The model prediction of Image classification is a list, and that if Object Detection is a dict
        if not isinstance(model_prediction, list) and not isinstance(model_prediction, dict):
            raise ValueError("Failed to construct FeedbackInput. ModelPrediction param must be a list or a dict.")
        if not model_prediction:
            raise ValueError("Failed to construct FeedbackInput. ModelPrediction param cannot be empty.")
        self._model_prediction = model_prediction

        self._metadata = context.get(PARAM_METADATA)
        if self._metadata:
            if not isinstance(self._metadata, dict):
                raise ValueError("Failed to construct FeedbackInput. Metadata param must be a dict.")
        else:
            self._metadata = {}

        self._prediction_topic = context.get(PARAM_PREDICTION_TOPIC)
        self._status_topic = context.get(PARAM_STATUS_TOPIC)

        self._filename = context.get(PARAM_FILENAME)

        self._model_input = model_input

    @property
    def config_id(self):
        return self._config_id.id

    @property
    def model_prediction(self):
        return self._model_prediction

    @property
    def metadata(self):
        return self._metadata

    @property
    def prediction_topic(self):
        return self._prediction_topic

    @property
    def filename(self):
        return self._filename

    @property
    def status_topic(self):
        return self._status_topic

    @property
    def model_input(self):
        return self._model_input
